/* charset.c - Maintain a character set for a crossword puzzle
 *
 * Copyright 2021 Federico Mena Quintero <federico@gnome.org>
 * Copyright 2023 Jonathan Blandford <jrb@gnome.org>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * SPDX-License-Identifier: (LGPL-2.1-or-later OR MIT)
 */

#include <locale.h>
#include "ipuz-charset.h"
#include "ipuz-charset-lang.h"

struct _IPuzCharset
{
  grefcount ref_count;

  /* Keys are GUINT_TO_POINTER(gunichar), values are a count of the
   * number of instances in the seed strings */
  GTree *tree;
};


G_DEFINE_BOXED_TYPE(IPuzCharset, ipuz_charset, ipuz_charset_ref, ipuz_charset_unref);


static gint
ipuz_charset_compare_cb (gconstpointer a,
                         gconstpointer b)
{
  gunichar ca = GPOINTER_TO_UINT (a);
  gunichar cb = GPOINTER_TO_UINT (b);

  if (ca < cb)
    return -1;
  else if (ca > cb)
    return 1;
  else
    return 0;
}

/**
 * ipuz_charset_new:
 *
 * Returns: an empty character set.  Use `ipuz_charset_add_text()` to populate it.
 */
IPuzCharset *
ipuz_charset_new (void)
{
  IPuzCharset *charset;

  charset = g_new0 (IPuzCharset, 1);
  g_ref_count_init (&charset->ref_count);
  charset->tree = g_tree_new (ipuz_charset_compare_cb);

  return charset;
}

/**
 * ipuz_charset_new:
 * @text: the text to bas a new IPuzCharset on, or NULL
 *
 * Returns: a new character set populated by @text
 */
IPuzCharset *
ipuz_charset_new_from_text (const char *text)
{
  IPuzCharset *charset;

  charset = ipuz_charset_new ();
  ipuz_charset_add_text (charset, text);

  return charset;
}

/**
 * ipuz_charset_new_for_language:
 * @lang: A language code, such as "en" or "es"
 *
 * Creates a charset with a list of all characters in common use in
 * crosswords for @lang's alphabet. @lang should be a country code,
 * but can be a fully-qualified locale (such as from the $LANG
 * environment variable). In that instance the remainder of the string
 * is ignored, as we don't consider regional distinctions when
 * determining a default alphabet. Along those lines, a lang of "C"
 * will return the English alphabet.
 *
 * Note that this returns the common alphabet of letters for a
 * language and will not include digraphs as independent
 * characters. As examples, Dutch will not include a separate 'ĳ'
 * digraph, despite the prevelance of "IJ" in Dutch puzzles.
 *
 * Returns: a newly allocated @IPuzCharset, or NULL
 **/
IPuzCharset *
ipuz_charset_new_for_language (const char *lang)
{
  g_return_val_if_fail (lang != NULL, NULL);

  if (!g_strcmp0 ("C", lang))
    return ipuz_charset_new_from_text (CHARSET_EN_ALPHABET);
  else if (!g_ascii_strncasecmp ("EN", lang, 2))
    return ipuz_charset_new_from_text (CHARSET_EN_ALPHABET);
  else if (!g_ascii_strncasecmp ("ES", lang, 2))
    return ipuz_charset_new_from_text (CHARSET_ES_ALPHABET);
  else if (!g_ascii_strncasecmp ("NL", lang, 2))
    return ipuz_charset_new_from_text (CHARSET_NL_ALPHABET);
  else if (!g_ascii_strncasecmp ("IT", lang, 2))
    return ipuz_charset_new_from_text (CHARSET_IT_ALPHABET);

  return NULL;
}

/**
 * ipuz_charset_ref:
 * @charset: the character set to ref
 *
 * Refs the character set.
 *
 * Returns: @charset. This can be used to chain calls or ref on return.
 */
IPuzCharset *
ipuz_charset_ref (IPuzCharset *charset)
{
  g_return_val_if_fail (charset != NULL, NULL);

  g_ref_count_inc (&charset->ref_count);

  return charset;
}

/**
 * ipuz_charset_unref:
 * @charset: the character set to unref
 *
 * Frees a character set.
 */
void
ipuz_charset_unref (IPuzCharset *charset)
{
  g_return_if_fail (charset != NULL);

  if (!g_ref_count_dec (&charset->ref_count))
    return;

  /* Free */
  g_tree_unref (charset->tree);
  g_free (charset);
}

gboolean
ipuz_charset_compare          (IPuzCharset       *charset_a,
                               IPuzCharset       *charset_b)
{
  IPuzCharsetIter *iter_a;
  IPuzCharsetIter *iter_b;

  if (charset_a == NULL && charset_b == NULL)
    return TRUE;

  if (charset_a == NULL || charset_b == NULL)
    return FALSE;

  iter_a = ipuz_charset_iter_first (charset_a);
  iter_b = ipuz_charset_iter_first (charset_b);

  while (iter_a && iter_b)
    {
      IPuzCharsetIterValue value_a;
      IPuzCharsetIterValue value_b;

      value_a = ipuz_charset_iter_get_value (iter_a);
      value_b = ipuz_charset_iter_get_value (iter_b);

      if (value_a.c != value_b.c)
        return FALSE;

      if (value_a.count != value_b.count)
        return FALSE;

      iter_a = ipuz_charset_iter_next (iter_a);
      iter_b = ipuz_charset_iter_next (iter_b);
    }
  if (iter_a == NULL && iter_b == NULL)
    return TRUE;

  return FALSE;
}

/**
 * ipuz_charset_add_text:
 * @charset: the character set to populate.
 * @text: string with characters to add to the charset.
 *
 * Adds each unicode code point from @str into the @charset.
 */
void
ipuz_charset_add_text (IPuzCharset *charset,
                       const char  *text)
{
  const char *p;

  g_return_if_fail (charset != NULL);

  for (p = text; p && p[0]; p = g_utf8_next_char (p))
    {
      gunichar c;

      c = g_utf8_get_char (p);
      ipuz_charset_add_character (charset, c);
    }
}

/**
 * ipuz_charset_add_character:
 * @charset: the character set to extend.
 * @c: a unicode character to add to the charset
 *
 * Adds @c to the @charset.
 **/
void
ipuz_charset_add_character (IPuzCharset *charset,
                            gunichar     c)
{
  GTreeNode *node;
      gpointer key;

  g_return_if_fail (charset != NULL);

  key = GUINT_TO_POINTER(c);
  node = g_tree_lookup_node (charset->tree, key);

  if (node == NULL)
    {
      g_tree_insert (charset->tree, key, GUINT_TO_POINTER (1));
    }
  else
    {
      guint count = GPOINTER_TO_UINT (g_tree_node_value (node));
      g_tree_replace (charset->tree, key, GUINT_TO_POINTER (count + 1));
    }
}

typedef struct {
  gunichar needle;
  gsize index;
} IndexTraverse;

static gboolean
index_traverse_cb (gpointer key,
                   gpointer value,
                   gpointer data)
{
  gunichar ch = GPOINTER_TO_UINT (key);
  IndexTraverse *traverse = data;

  if (ch == traverse->needle)
    {
      return TRUE;
    }
  else
    {
      traverse->index += 1;
      return FALSE;
    }
}

/**
 * ipuz_charset_get_char_index:
 * @charset: the character set to search in.
 * @c: character to search for.
 *
 * Returns the index of the character @c in the @charset, or -1 if it does not exist.
 */
gint
ipuz_charset_get_char_index (const IPuzCharset *charset,
                             gunichar           c)
{
  IndexTraverse traverse = {
    .needle = c,
    .index = 0,
  };

  g_tree_foreach (charset->tree, index_traverse_cb, &traverse);

  if (traverse.index < ipuz_charset_get_n_chars (charset))
    {
      return (gint) traverse.index;
    }
  else
    {
      return -1;
    }
}

/**
 * ipuz_charset_get_char_count:
 * @charset: the character set to lookup in
 * @c: a character
 *
 *
 *
 * Returns: the number of instances of @c in the text
 **/
guint
ipuz_charset_get_char_count (const IPuzCharset *charset,
                             gunichar           c)
{
  g_return_val_if_fail (charset != NULL, 0);

  return GPOINTER_TO_UINT (g_tree_lookup (charset->tree, GUINT_TO_POINTER (c)));
}

/**
 * ipuz_charset_get_n_chars:
 * @charset: the character set to query.
 *
 * Returns: the number of different types of characters stored in the
 * @charset. This is a constant-time operation.
 */
gsize
ipuz_charset_get_n_chars (const IPuzCharset *charset)
{
  return g_tree_nnodes (charset->tree);
}

/**
 * ipuz_charset_get_size:
 * @charset: the character set to query
 *
 * Returns the total number of characters stored in @charset. Unlike
 * ipuz_charset_get_n_chars() which returns the number of types of
 * characters, this returns the count of characters.
 *
 * Returns: Total number of characters
 **/
gsize
ipuz_charset_get_size (const IPuzCharset *charset)
{
  IPuzCharsetIter *iter;
  gsize size = 0;

  g_return_val_if_fail (charset != NULL, 0);

  for (iter = ipuz_charset_iter_first (charset);
       iter;
       iter = ipuz_charset_iter_next (iter))
    {
      IPuzCharsetIterValue value;

      value = ipuz_charset_iter_get_value (iter);
      size += value.count;
    }

  return size;
}

typedef struct {
  GString *string;
} SerializeTraverse;

static gboolean
serialize_traverse_cb (gpointer key,
                       gpointer value,
                       gpointer data)
{
  gunichar ch = GPOINTER_TO_UINT (key);
  SerializeTraverse *traverse = data;

  g_string_append_unichar (traverse->string, ch);
  return FALSE;
}

/**
 * ipuz_charset_iter_first:
 * @charset: A @IPuzCharset
 *
 * Gets an iteratior for querying the charset. Note, this iterator is
 * invalidated if @charset is changed.
 *
 * Returns: The first iterator
 **/
IPuzCharsetIter *
ipuz_charset_iter_first (const IPuzCharset *charset)
{
  g_return_val_if_fail (charset != NULL, NULL);

  return (IPuzCharsetIter *) g_tree_node_first (charset->tree);
}

/**
 * ipuz_charset_iter_next:
 * @iter: A @IPuzCharsetIter
 *
 * Returns the next iter in the charset sequentially. If we've reached
 * the end of the set, then return %NULL.
 *
 * Returns: the next @IPuzCharsetIter, or %NULL
 **/
IPuzCharsetIter *
ipuz_charset_iter_next (IPuzCharsetIter *iter)
{
  return (IPuzCharsetIter *) g_tree_node_next ((GTreeNode *)iter);
}

IPuzCharsetIterValue
ipuz_charset_iter_get_value (IPuzCharsetIter *iter)
{
  IPuzCharsetIterValue value;

  value.c = GPOINTER_TO_UINT (g_tree_node_key ((GTreeNode *)iter));
  value.count = GPOINTER_TO_UINT (g_tree_node_value ((GTreeNode *)iter));

  return value;
}

/**
 * ipuz_charset_serialize:
 * @charset: the character set to serialize.
 *
 * Concatenates all the unique characters stored in a @charset, in the order in which they
 * would be returned by ipuz_charset_get_char_index().
 *
 * Returns: a string with all the characters from the character set.
 */
char *
ipuz_charset_serialize (const IPuzCharset *charset)
{
  SerializeTraverse traverse = {
    .string = g_string_new (NULL),
  };

  g_tree_foreach (charset->tree, serialize_traverse_cb, &traverse);
  return g_string_free (traverse.string, FALSE);
}

/**
 * ipuz_charset_deserialize:
 * @str: String serialization of a character set, as returned by ipuz_charset_serialize().
 *
 * Creates a new character set by deserializing from a string.
 *
 * Returns: a new character set.
 */
IPuzCharset *
ipuz_charset_deserialize (const char *str)
{
  IPuzCharset *charset = ipuz_charset_new ();
  ipuz_charset_add_text (charset, str);
  return charset;
}

static gboolean
ipuz_charset_clone_foreach (gpointer key,
                            gpointer value,
                            gpointer data)
{
  IPuzCharset *charset = (IPuzCharset *)data;

  g_tree_insert (charset->tree, key, value);

  return FALSE;
}

IPuzCharset*
ipuz_charset_clone (const IPuzCharset *ipuz_charset)
{
  IPuzCharset *charset;

  charset = ipuz_charset_new ();

  g_tree_foreach (ipuz_charset->tree, (GTraverseFunc)ipuz_charset_clone_foreach, charset);

  return charset;
}

gboolean
ipuz_charset_contains_text (const IPuzCharset *ipuz_charset,
                            const char        *text)
{
  const char *p;
  g_autoptr (IPuzCharset) charset = NULL;

  charset = ipuz_charset_clone (ipuz_charset);

  for (p = text; p[0]; p = g_utf8_next_char (p))
    {
      GTreeNode *node;
      gunichar c;
      gpointer key;

      c = g_utf8_get_char (p);

      key = GUINT_TO_POINTER(c);
      node = g_tree_lookup_node (charset->tree, key);

      if (node != NULL)
        {
          guint count = GPOINTER_TO_UINT (g_tree_node_value (node));

          if (count == 1)
            g_tree_remove (charset->tree, key);
          else
            g_tree_replace (charset->tree, key, GUINT_TO_POINTER (count - 1));
        }
      else
	{
	  return FALSE;
	}
    }

  return TRUE;
}

gboolean
ipuz_charset_remove_text (IPuzCharset *ipuz_charset,
                          const char  *text)
{
  const char *p;

  if (!ipuz_charset_contains_text (ipuz_charset, text))
    return FALSE;

  for (p = text; p[0]; p = g_utf8_next_char (p))
    {
      GTreeNode *node;
      gunichar c;
      gpointer key;

      c = g_utf8_get_char (p);

      key = GUINT_TO_POINTER(c);
      node = g_tree_lookup_node (ipuz_charset->tree, key);

      if (node != NULL)
        {
          guint count = GPOINTER_TO_UINT (g_tree_node_value (node));

          if (count == 1)
            g_tree_remove (ipuz_charset->tree, key);
          else
            g_tree_replace (ipuz_charset->tree, key, GUINT_TO_POINTER (count - 1));
        }
    }

  return TRUE;
}
